@@ -42,6 +42,13 @@
 extern int mifare_read(uint32_t adapter_idx, uint32_t target_idx,
 		near_tag_io_cb cb, enum near_tag_sub_type tgt_subtype);
 
+extern int mifare_check_presence(uint32_t adapter_idx, uint32_t target_idx,
+			near_tag_io_cb cb, enum near_tag_sub_type tgt_subtype);
+
+extern int mifare_write(uint32_t adapter_idx, uint32_t target_idx,
+			struct near_ndef_message *ndef,
+			near_tag_io_cb cb, enum near_tag_sub_type tgt_subtype);
+
 #define CMD_READ         0x30
 #define CMD_READ_SIZE    0x02
 
@@ -97,10 +104,17 @@ struct type2_cc {
 	uint8_t read_write;
 };
 
-static void t2_cookie_release(struct t2_cookie *cookie)
+static int t2_cookie_release(int err, void *data)
 {
+	struct t2_cookie *cookie = data;
+
+	DBG("%p", cookie);
+
 	if (cookie == NULL)
-		return;
+		return err;
+
+	if (cookie->cb != NULL)
+		cookie->cb(cookie->adapter_idx, cookie->target_idx, err);
 
 	if (cookie->ndef)
 		g_free(cookie->ndef->data);
@@ -108,6 +122,8 @@ static void t2_cookie_release(struct t2_cookie *cookie)
 	g_free(cookie->ndef);
 	g_free(cookie);
 	cookie = NULL;
+
+	return err;
 }
 
 static int data_recv(uint8_t *resp, int length, void *data)
@@ -162,8 +178,8 @@ static int data_recv(uint8_t *resp, int length, void *data)
 	DBG("adapter %d", adapter_idx);
 
 	return near_adapter_send(adapter_idx,
-				(uint8_t *)&cmd, CMD_READ_SIZE,
-					data_recv, tag);
+				(uint8_t *) &cmd, CMD_READ_SIZE,
+					data_recv, tag, NULL);
 }
 
 static int data_read(struct type2_tag *tag)
@@ -181,8 +197,8 @@ static int data_read(struct type2_tag *tag)
 	adapter_idx = near_tag_get_adapter_idx(tag->tag);
 
 	return near_adapter_send(adapter_idx,
-				(uint8_t *)&cmd, CMD_READ_SIZE,
-					data_recv, tag);
+					(uint8_t *) &cmd, CMD_READ_SIZE,
+					data_recv, tag, NULL);
 }
 
 static int meta_recv(uint8_t *resp, int length, void *data)
@@ -197,12 +213,12 @@ static int meta_recv(uint8_t *resp, int length, void *data)
 
 	if (length < 0) {
 		err = length;
-		goto out;
+		goto out_err;
 	}
 
 	if (resp[0] != 0) {
 		err = -EIO;
-		goto out;
+		goto out_err;
 	}
 
 	cc = TAG_DATA_CC(resp + NFC_HEADER_SIZE);
@@ -213,18 +229,18 @@ static int meta_recv(uint8_t *resp, int length, void *data)
 			TYPE2_DATA_SIZE_48 << 3));
 
 	if (err < 0)
-		goto out;
+		goto out_err;
 
 	tag = near_tag_get_tag(cookie->adapter_idx, cookie->target_idx);
 	if (tag == NULL) {
 		err = -ENOMEM;
-		goto out;
+		goto out_err;
 	}
 
 	t2_tag = g_try_malloc0(sizeof(struct type2_tag));
 	if (t2_tag == NULL) {
 		err = -ENOMEM;
-		goto out;
+		goto out_err;
 	}
 
 	t2_tag->adapter_idx = cookie->adapter_idx;
@@ -248,19 +264,20 @@ static int meta_recv(uint8_t *resp, int length, void *data)
 
 	err = data_read(t2_tag);
 	if (err < 0)
-		goto out;
+		goto out_tag;
 
-	t2_cookie_release(cookie);
+	/*
+	 * As reading isn't complete,
+	 * callback shouldn't be called while freeing the cookie
+	 */
+	cookie->cb = NULL;
+	return t2_cookie_release(err, cookie);
 
-	return 0;
+out_tag:
+	g_free(t2_tag);
 
-out:
-	if (err < 0 && cookie->cb)
-		cookie->cb(cookie->adapter_idx, cookie->target_idx, err);
-
-	t2_cookie_release(cookie);
-
-	return err;
+out_err:
+	return t2_cookie_release(err, cookie);
 }
 
 static int nfctype2_read_meta(uint32_t adapter_idx, uint32_t target_idx,
@@ -268,25 +285,27 @@ static int nfctype2_read_meta(uint32_t adapter_idx, uint32_t target_idx,
 {
 	struct type2_cmd cmd;
 	struct t2_cookie *cookie;
-	
+
 	DBG("");
 
 	cmd.cmd = CMD_READ;
 	cmd.block = META_BLOCK_START;
 
 	cookie = g_try_malloc0(sizeof(struct t2_cookie));
+	if (cookie == NULL)
+		return -ENOMEM;
+
 	cookie->adapter_idx = adapter_idx;
 	cookie->target_idx = target_idx;
 	cookie->cb = cb;
 
-	return near_adapter_send(adapter_idx, (uint8_t *)&cmd, CMD_READ_SIZE,
-							meta_recv, cookie);
+	return near_adapter_send(adapter_idx, (uint8_t *) &cmd, CMD_READ_SIZE,
+					meta_recv, cookie, t2_cookie_release);
 }
 
 static int nfctype2_read(uint32_t adapter_idx,
 				uint32_t target_idx, near_tag_io_cb cb)
 {
-	int err;
 	enum near_tag_sub_type tgt_subtype;
 
 	DBG("");
@@ -295,23 +314,18 @@ static int nfctype2_read(uint32_t adapter_idx,
 
 	switch (tgt_subtype) {
 	case NEAR_TAG_NFC_T2_MIFARE_ULTRALIGHT:
-		err = nfctype2_read_meta(adapter_idx, target_idx, cb);
-		break;
+		return nfctype2_read_meta(adapter_idx, target_idx, cb);
 
 	/* Specific Mifare read access */
 	case NEAR_TAG_NFC_T2_MIFARE_CLASSIC_1K:
 	case NEAR_TAG_NFC_T2_MIFARE_CLASSIC_4K:
-		err= mifare_read( adapter_idx, target_idx,
+		return mifare_read(adapter_idx, target_idx,
 			cb, tgt_subtype);
-		break;
 
 	default:
-		DBG("Unknown TAG Type 2 subtype (%d)", tgt_subtype);
-		err = -1;
-		break;
+		DBG("Unknown Tag Type 2 subtype %d", tgt_subtype);
+		return -1;
 	}
-
-	return err;
 }
 
 static int data_write_resp(uint8_t *resp, int length, void *data)
@@ -324,18 +338,13 @@ static int data_write_resp(uint8_t *resp, int length, void *data)
 
 	if (length < 0 || resp[0] != 0) {
 		err = -EIO;
-		goto out;
+		goto out_err;
 	}
 
 	if (cookie->ndef->offset > cookie->ndef->length) {
 		DBG("Done writing");
 
-		if (cookie->cb)
-			cookie->cb(cookie->adapter_idx, cookie->target_idx, 0);
-
-		t2_cookie_release(cookie);
-
-		return 0;
+		return t2_cookie_release(0, cookie);
 	}
 
 	cmd.cmd = CMD_WRITE;
@@ -353,22 +362,12 @@ static int data_write_resp(uint8_t *resp, int length, void *data)
 		cookie->ndef->offset = cookie->ndef->length + 1;
 	}
 
-	err = near_adapter_send(cookie->adapter_idx, (uint8_t *)&cmd,
-					sizeof(cmd), data_write_resp, cookie);
+	return near_adapter_send(cookie->adapter_idx, (uint8_t *) &cmd,
+					sizeof(cmd), data_write_resp, cookie,
+					t2_cookie_release);
 
-
-	if (err < 0)
-		goto out;
-
-	return 0;
-
-out:
-	if (err < 0 && cookie->cb)
-		cookie->cb(cookie->adapter_idx, cookie->target_idx, err);
-
-	t2_cookie_release(cookie);
-
-	return err;
+out_err:
+	return t2_cookie_release(err, cookie);
 }
 
 static int data_write(uint32_t adapter_idx, uint32_t target_idx,
@@ -382,6 +381,13 @@ static int data_write(uint32_t adapter_idx, uint32_t target_idx,
 	DBG("");
 
 	cookie = g_try_malloc0(sizeof(struct t2_cookie));
+	if (cookie == NULL) {
+		err = -ENOMEM;
+		if (cb != NULL)
+			cb(adapter_idx, target_idx, err);
+		return err;
+	}
+
 	cookie->adapter_idx = adapter_idx;
 	cookie->target_idx = target_idx;
 	cookie->current_block = DATA_BLOCK_START;
@@ -394,18 +400,9 @@ static int data_write(uint32_t adapter_idx, uint32_t target_idx,
 	cookie->ndef->offset += BLOCK_SIZE;
 	cookie->current_block++;
 
-	err = near_adapter_send(cookie->adapter_idx, (uint8_t *)&cmd,
-					sizeof(cmd), data_write_resp, cookie);
-
-	if (err < 0)
-		goto out;
-
-	return 0;
-
-out:
-	t2_cookie_release(cookie);
-
-	return err;
+	return near_adapter_send(cookie->adapter_idx, (uint8_t *) &cmd,
+					sizeof(cmd), data_write_resp, cookie,
+					t2_cookie_release);
 }
 
 static int nfctype2_write(uint32_t adapter_idx, uint32_t target_idx,
@@ -414,33 +411,58 @@ static int nfctype2_write(uint32_t adapter_idx, uint32_t target_idx,
 {
 	struct near_tag *tag;
 	enum near_tag_sub_type tgt_subtype;
+	int err;
 
 	DBG("");
 
-	if (ndef == NULL || cb == NULL)
-		return -EINVAL;
+	if (ndef == NULL || cb == NULL) {
+		err = -EINVAL;
+		goto out_err;
+	}
 
 	tag = near_tag_get_tag(adapter_idx, target_idx);
-	if (tag == NULL)
-		return -EINVAL;
+	if (tag == NULL) {
+		err = -EINVAL;
+		goto out_err;
+	}
 
 	tgt_subtype = near_tag_get_subtype(adapter_idx, target_idx);
 
-	if (tgt_subtype != NEAR_TAG_NFC_T2_MIFARE_ULTRALIGHT) {
-		DBG("Unknown Tag Type 2 subtype (%d)", tgt_subtype);
-		return -1;
-	}
-
-	/* This check is valid for only static tags.
-	 * Max data length on Type 2 Tag including TLV's is NDEF_MAX_SIZE */
-	if (near_tag_get_memory_layout(tag) == NEAR_TAG_MEMORY_STATIC) {
-		if ((ndef->length + 3) > NDEF_MAX_SIZE) {
-			near_error("not enough space on tag");
-			return -ENOSPC;
+	switch (tgt_subtype) {
+	case NEAR_TAG_NFC_T2_MIFARE_ULTRALIGHT:
+		/*
+		 * This check is valid for only static tags.
+		 * Max data length on Type 2 Tag
+		 * including TLV's is NDEF_MAX_SIZE
+		 */
+		if (near_tag_get_memory_layout(tag) == NEAR_TAG_MEMORY_STATIC) {
+			if ((ndef->length + 3) > NDEF_MAX_SIZE) {
+				near_error("not enough space on tag");
+				err = -ENOSPC;
+				goto out_err;
+			}
 		}
+
+		return data_write(adapter_idx, target_idx, ndef, cb);
+
+	/* Specific Mifare write access */
+	case NEAR_TAG_NFC_T2_MIFARE_CLASSIC_1K:
+	case NEAR_TAG_NFC_T2_MIFARE_CLASSIC_4K:
+		return mifare_write(adapter_idx, target_idx, ndef,
+				cb, tgt_subtype);
+	default:
+		DBG("Unknown TAG Type 2 subtype %d", tgt_subtype);
+		err = -EINVAL;
+		goto out_err;
 	}
 
-	return data_write(adapter_idx, target_idx, ndef, cb);
+	return 0;
+
+out_err:
+	if (cb != NULL)
+		cb(adapter_idx, target_idx, err);
+
+	return err;
 }
 
 static int check_presence(uint8_t *resp, int length, void *data)
@@ -453,13 +475,7 @@ static int check_presence(uint8_t *resp, int length, void *data)
 	if (length < 0)
 		err = -EIO;
 
-	if (cookie->cb)
-		cookie->cb(cookie->adapter_idx,
-				cookie->target_idx, err);
-
-	t2_cookie_release(cookie);
-
-	return err;
+	return t2_cookie_release(err, cookie);
 }
 
 static int nfctype2_check_presence(uint32_t adapter_idx, uint32_t target_idx,
@@ -467,32 +483,40 @@ static int nfctype2_check_presence(uint32_t adapter_idx, uint32_t target_idx,
 {
 	struct type2_cmd cmd;
 	struct t2_cookie *cookie;
-	int err;
+	enum near_tag_sub_type tgt_subtype;
 
 	DBG("");
 
-	cmd.cmd = CMD_READ;
-	cmd.block = META_BLOCK_START;
+	tgt_subtype = near_tag_get_subtype(adapter_idx, target_idx);
 
-	cookie = g_try_malloc0(sizeof(struct t2_cookie));
-	if (cookie == NULL)
-		return -ENOMEM;
+	switch (tgt_subtype) {
+	case NEAR_TAG_NFC_T2_MIFARE_ULTRALIGHT:
+		cmd.cmd = CMD_READ;
+		cmd.block = META_BLOCK_START;
 
-	cookie->adapter_idx = adapter_idx;
-	cookie->target_idx = target_idx;
-	cookie->cb = cb;
+		cookie = g_try_malloc0(sizeof(struct t2_cookie));
+		if (cookie == NULL)
+			return -ENOMEM;
 
-	err = near_adapter_send(adapter_idx, (uint8_t *)&cmd, CMD_READ_SIZE,
-							check_presence, cookie);
-	if (err < 0)
-		goto out;
+		cookie->adapter_idx = adapter_idx;
+		cookie->target_idx = target_idx;
+		cookie->cb = cb;
 
-	return 0;
+		return near_adapter_send(adapter_idx, (uint8_t *) &cmd,
+					CMD_READ_SIZE, check_presence, cookie,
+					t2_cookie_release);
 
-out:
-	t2_cookie_release(cookie);
+	/* Specific Mifare check presence */
+	case NEAR_TAG_NFC_T2_MIFARE_CLASSIC_1K:
+	case NEAR_TAG_NFC_T2_MIFARE_CLASSIC_4K:
+		return mifare_check_presence(adapter_idx, target_idx,
+							cb, tgt_subtype);
 
-	return err;
+	default:
+		DBG("Unknown TAG Type 2 subtype %d", tgt_subtype);
+
+		return -1;
+	}
 }
 
 static int format_resp(uint8_t *resp, int length, void *data)
@@ -505,25 +529,20 @@ static int format_resp(uint8_t *resp, int length, void *data)
 
 	if (length < 0 || resp[0] != 0) {
 		err = -EIO;
-		goto out;
+		goto out_err;
 	}
 
 	tag = near_tag_get_tag(cookie->adapter_idx, cookie->target_idx);
 	if (tag == NULL) {
 		err = -EINVAL;
-		goto out;
+		goto out_err;
 	}
 
 	DBG("Done formatting");
 	near_tag_set_blank(tag, FALSE);
 
-out:
-	if (cookie->cb)
-		cookie->cb(cookie->adapter_idx, cookie->target_idx, err);
-
-	t2_cookie_release(cookie);
-
-	return err;
+out_err:
+	return t2_cookie_release(err, cookie);
 }
 
 static int nfctype2_format(uint32_t adapter_idx, uint32_t target_idx,
@@ -547,31 +566,44 @@ static int nfctype2_format(uint32_t adapter_idx, uint32_t target_idx,
 	tgt_subtype = near_tag_get_subtype(adapter_idx, target_idx);
 
 	if (tgt_subtype != NEAR_TAG_NFC_T2_MIFARE_ULTRALIGHT) {
-		DBG("Unknown Tag Type 2 subtype (%d)", tgt_subtype);
+		DBG("Unknown Tag Type 2 subtype %d", tgt_subtype);
 		return -1;
 	}
 
 	t2_cc = g_try_malloc0(sizeof(struct type2_cc));
+	cc_ndef = g_try_malloc0(sizeof(struct near_ndef_message));
+	cookie = g_try_malloc0(sizeof(struct t2_cookie));
+
+	if (t2_cc == NULL || cc_ndef == NULL || cookie == NULL) {
+		err = -ENOMEM;
+		goto out_err;
+	}
+
 	t2_cc->magic = TYPE2_MAGIC;
 	t2_cc->version = TYPE2_TAG_VER_1_0;
 	t2_cc->mem_size = TYPE2_DATA_SIZE_48;
 	t2_cc->read_write = TYPE2_READWRITE_ACCESS;
 
-	cc_ndef = g_try_malloc0(sizeof(struct near_ndef_message));
-	cookie = g_try_malloc0(sizeof(struct t2_cookie));
 	cookie->adapter_idx = adapter_idx;
 	cookie->target_idx = target_idx;
 	cookie->current_block = CC_BLOCK_START;
 	cookie->ndef = cc_ndef;
-	cookie->ndef->data = (uint8_t *)t2_cc;
+	cookie->ndef->data = (uint8_t *) t2_cc;
 	cookie->cb = cb;
 
 	cmd.cmd = CMD_WRITE;
 	cmd.block = CC_BLOCK_START;
-	memcpy(cmd.data, (uint8_t *)t2_cc, BLOCK_SIZE);
+	memcpy(cmd.data, (uint8_t *) t2_cc, BLOCK_SIZE);
 
-	err = near_adapter_send(cookie->adapter_idx, (uint8_t *)&cmd,
-					sizeof(cmd), format_resp, cookie);
+	err = near_adapter_send(cookie->adapter_idx, (uint8_t *) &cmd,
+				sizeof(cmd), format_resp, cookie, NULL);
+
+out_err:
+	if (err < 0) {
+		g_free(t2_cc);
+		g_free(cc_ndef);
+		g_free(cookie);
+	}
 
 	return err;
 }
@@ -601,4 +633,3 @@ static void nfctype2_exit(void)
 
 NEAR_PLUGIN_DEFINE(nfctype2, "NFC Forum Type 2 tags support", VERSION,
 			NEAR_PLUGIN_PRIORITY_HIGH, nfctype2_init, nfctype2_exit)
-
