@@ -42,6 +42,12 @@ static DBusConnection *connection = NULL;
 
 static GHashTable *adapter_hash;
 
+enum near_adapter_rf_mode {
+	NEAR_ADAPTER_RF_MODE_IDLE      = 0,
+	NEAR_ADAPTER_RF_MODE_INITIATOR = 1,
+	NEAR_ADAPTER_RF_MODE_TARGET    = 2
+};
+
 #define NEAR_ADAPTER_MODE_INITIATOR 0x1
 #define NEAR_ADAPTER_MODE_TARGET    0x2
 #define NEAR_ADAPTER_MODE_DUAL      0x3
@@ -53,7 +59,7 @@ struct near_adapter {
 	uint32_t idx;
 	uint32_t protocols;
 	uint32_t poll_mode;
-	uint32_t rf_mode;
+	enum near_adapter_rf_mode rf_mode;
 
 	near_bool_t powered;
 	near_bool_t polling;
@@ -97,6 +103,8 @@ static void free_adapter(gpointer data)
 
 	g_free(adapter->name);
 	g_free(adapter->path);
+	g_hash_table_destroy(adapter->tags);
+	g_hash_table_destroy(adapter->devices);
 	g_free(adapter);
 }
 
@@ -114,6 +122,20 @@ static void free_device(gpointer data)
 	__near_device_remove(device);
 }
 
+static char *rf_mode_to_string(struct near_adapter *adapter)
+{
+	switch (adapter->rf_mode) {
+	case NEAR_ADAPTER_RF_MODE_IDLE:
+		return "Idle";
+	case NEAR_ADAPTER_RF_MODE_INITIATOR:
+		return "Initiator";
+	case NEAR_ADAPTER_RF_MODE_TARGET:
+		return "Target";
+	}
+
+	return NULL;
+}
+
 static void polling_changed(struct near_adapter *adapter)
 {
 
@@ -122,6 +144,18 @@ static void polling_changed(struct near_adapter *adapter)
 					DBUS_TYPE_BOOLEAN, &adapter->polling);
 }
 
+static void rf_mode_changed(struct near_adapter *adapter)
+{
+	const char *rf_mode = rf_mode_to_string(adapter);
+
+	if (rf_mode == NULL)
+		return;
+
+	near_dbus_property_changed_basic(adapter->path,
+					NFC_ADAPTER_INTERFACE, "Mode",
+					DBUS_TYPE_STRING, &rf_mode);
+}
+
 static int adapter_start_poll(struct near_adapter *adapter)
 {
 	int err;
@@ -306,6 +340,7 @@ static DBusMessage *get_properties(DBusConnection *conn,
 					DBusMessage *msg, void *data)
 {
 	struct near_adapter *adapter = data;
+	const char *rf_mode;
 	DBusMessage *reply;
 	DBusMessageIter array, dict;
 
@@ -325,6 +360,11 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	near_dbus_dict_append_basic(&dict, "Polling",
 				    DBUS_TYPE_BOOLEAN, &adapter->polling);
 
+	rf_mode = rf_mode_to_string(adapter);
+	if (rf_mode != NULL)
+		near_dbus_dict_append_basic(&dict, "Mode",
+						DBUS_TYPE_STRING, &rf_mode);
+
 	near_dbus_dict_append_array(&dict, "Protocols",
 				DBUS_TYPE_STRING, append_protocols, adapter);
 
@@ -553,7 +593,7 @@ static const GDBusSignalTable adapter_signals[] = {
 	{ }
 };
 
-struct near_adapter * __near_adapter_create(uint32_t idx,
+struct near_adapter *__near_adapter_create(uint32_t idx,
 		const char *name, uint32_t protocols, near_bool_t powered)
 {
 	struct near_adapter *adapter;
@@ -802,6 +842,9 @@ int __near_adapter_add_target(uint32_t idx, uint32_t target_idx,
 	adapter->polling = FALSE;
 	polling_changed(adapter);
 
+	adapter->rf_mode = NEAR_ADAPTER_RF_MODE_INITIATOR;
+	rf_mode_changed(adapter);
+
 	if (protocols & NFC_PROTO_NFC_DEP_MASK)
 		return adapter_add_device(adapter, target_idx,
 						nfcid, nfcid_len);
@@ -820,6 +863,9 @@ int __near_adapter_remove_target(uint32_t idx, uint32_t target_idx)
 	if (adapter == NULL)
 		return -ENODEV;
 
+	adapter->rf_mode = NEAR_ADAPTER_RF_MODE_IDLE;
+	rf_mode_changed(adapter);
+
 	if (g_hash_table_remove(adapter->tags,
 			GINT_TO_POINTER(target_idx)) == TRUE) {
 		__near_adapter_tags_changed(idx);
@@ -840,6 +886,7 @@ int __near_adapter_remove_target(uint32_t idx, uint32_t target_idx)
 int __near_adapter_add_device(uint32_t idx, uint8_t *nfcid, uint8_t nfcid_len)
 {
 	struct near_adapter *adapter;
+	int ret;
 
 	DBG("idx %d", idx);
 
@@ -849,11 +896,17 @@ int __near_adapter_add_device(uint32_t idx, uint8_t *nfcid, uint8_t nfcid_len)
 
 	adapter->polling = FALSE;
 	adapter->dep_up = TRUE;
+	adapter->rf_mode = NEAR_ADAPTER_RF_MODE_TARGET;
 	polling_changed(adapter);
+	rf_mode_changed(adapter);
+
+	ret = adapter_add_device(adapter, 0, nfcid, nfcid_len);
+	if (ret < 0)
+		return ret;
 
 	__near_adapter_devices_changed(idx);
 
-	return adapter_add_device(adapter, 0, nfcid, nfcid_len);
+	return 0;
 }
 
 int __near_adapter_remove_device(uint32_t idx)
@@ -871,6 +924,8 @@ int __near_adapter_remove_device(uint32_t idx)
 			GINT_TO_POINTER(device_idx)) == FALSE)
 		return 0;
 
+	adapter->rf_mode = NEAR_ADAPTER_RF_MODE_IDLE;
+	rf_mode_changed(adapter);
 	__near_adapter_devices_changed(idx);
 
 	adapter->dep_up = FALSE;
