@@ -49,16 +49,6 @@ static inline void nl_socket_free(struct nl_sock *h)
 	nl_handle_destroy(h);
 }
 
-static inline int __genl_ctrl_alloc_cache(struct nl_sock *h, struct nl_cache **cache)
-{
-	struct nl_cache *tmp = genl_ctrl_alloc_cache(h);
-	if (!tmp)
-		return -ENOMEM;
-	*cache = tmp;
-	return 0;
-}
-#define genl_ctrl_alloc_cache __genl_ctrl_alloc_cache
-
 #define NLE_MISSING_ATTR	14
 
 static inline void __nl_perror(int error, const char *s)
@@ -69,9 +59,9 @@ static inline void __nl_perror(int error, const char *s)
 #endif
 
 struct nlnfc_state {
-	struct nl_sock *nl_sock;
-	struct nl_cache *nl_cache;
-	struct genl_family *nlnfc;
+	struct nl_sock *cmd_sock;
+	struct nl_sock *event_sock;
+	int nfc_id;
 	int mcid;
 };
 
@@ -150,7 +140,6 @@ static int nl_send_msg(struct nl_sock *sock, struct nl_msg *msg,
 	return err;
 }
 
-
 static int get_devices_handler(struct nl_msg *n, void *arg)
 {
 	struct nlmsghdr *nlh = nlmsg_hdr(n);
@@ -170,7 +159,6 @@ static int get_devices_handler(struct nl_msg *n, void *arg)
 		return NL_STOP;
 	}
 
-
 	idx = nla_get_u32(attrs[NFC_ATTR_DEVICE_INDEX]);
 	name = nla_get_string(attrs[NFC_ATTR_DEVICE_NAME]);
 	protocols = nla_get_u32(attrs[NFC_ATTR_PROTOCOLS]);
@@ -189,27 +177,25 @@ int __near_netlink_get_adapters(void)
 {
 	struct nl_msg *msg;
 	void *hdr;
-	int err, family;
+	int err;
 
 	DBG("");
 
-	if (nfc_state == NULL || nfc_state->nlnfc == NULL)
+	if (nfc_state == NULL || nfc_state->nfc_id < 0)
 		return -ENODEV;
 
 	msg = nlmsg_alloc();
 	if (msg == NULL)
 		return -ENOMEM;
 
-	family = genl_family_get_id(nfc_state->nlnfc);
-
-	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, family, 0,
+	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, nfc_state->nfc_id, 0,
 			  NLM_F_DUMP, NFC_CMD_GET_DEVICE, NFC_GENL_VERSION);
 	if (hdr == NULL) {
 		err = -EINVAL;
 		goto out;
 	}
 
-	err = nl_send_msg(nfc_state->nl_sock, msg, get_devices_handler, NULL);
+	err = nl_send_msg(nfc_state->cmd_sock, msg, get_devices_handler, NULL);
 
 out:
 	nlmsg_free(msg);
@@ -217,31 +203,37 @@ out:
 	return err;
 }
 
-int __near_netlink_start_poll(int idx, uint32_t protocols)
+int __near_netlink_start_poll(int idx,
+				uint32_t im_protocols, uint32_t tm_protocols)
 {
 	struct nl_msg *msg;
 	void *hdr;
-	int family, err = 0;
+	int err;
 
-	DBG("");
+	DBG("IM protos 0x%x TM protos 0x%x", im_protocols, tm_protocols);
 
 	msg = nlmsg_alloc();
 	if (msg == NULL)
 		return -ENOMEM;
 
-	family = genl_family_get_id(nfc_state->nlnfc);
-
-	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, family, 0,
+	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, nfc_state->nfc_id, 0,
 			NLM_F_REQUEST, NFC_CMD_START_POLL, NFC_GENL_VERSION);
 	if (hdr == NULL) {
 		err = -EINVAL;
 		goto nla_put_failure;
 	}
 
-	NLA_PUT_U32(msg, NFC_ATTR_DEVICE_INDEX, idx);
-	NLA_PUT_U32(msg, NFC_ATTR_PROTOCOLS, protocols);
+	err = -EMSGSIZE;
 
-	err = nl_send_msg(nfc_state->nl_sock, msg, NULL, NULL);
+	NLA_PUT_U32(msg, NFC_ATTR_DEVICE_INDEX, idx);
+	if (im_protocols != 0) {
+		NLA_PUT_U32(msg, NFC_ATTR_IM_PROTOCOLS, im_protocols);
+		NLA_PUT_U32(msg, NFC_ATTR_PROTOCOLS, im_protocols);
+	}
+	if (tm_protocols != 0)
+		NLA_PUT_U32(msg, NFC_ATTR_TM_PROTOCOLS, tm_protocols);
+
+	err = nl_send_msg(nfc_state->cmd_sock, msg, NULL, NULL);
 
 nla_put_failure:
 	nlmsg_free(msg);
@@ -249,12 +241,11 @@ nla_put_failure:
 	return err;
 }
 
-
 int __near_netlink_stop_poll(int idx)
 {
 	struct nl_msg *msg;
 	void *hdr;
-	int family, err = 0;
+	int err;
 
 	DBG("");
 
@@ -262,18 +253,18 @@ int __near_netlink_stop_poll(int idx)
 	if (msg == NULL)
 		return -ENOMEM;
 
-	family = genl_family_get_id(nfc_state->nlnfc);
-
-	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, family, 0,
+	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, nfc_state->nfc_id, 0,
 			NLM_F_REQUEST, NFC_CMD_STOP_POLL, NFC_GENL_VERSION);
 	if (hdr == NULL) {
 		err = -EINVAL;
 		goto nla_put_failure;
 	}
 
+	err = -EMSGSIZE;
+
 	NLA_PUT_U32(msg, NFC_ATTR_DEVICE_INDEX, idx);
 
-	err = nl_send_msg(nfc_state->nl_sock, msg, NULL, NULL);
+	err = nl_send_msg(nfc_state->cmd_sock, msg, NULL, NULL);
 
 nla_put_failure:
 	nlmsg_free(msg);
@@ -286,7 +277,7 @@ int __near_netlink_dep_link_up(uint32_t idx, uint32_t target_idx,
 {
 	struct nl_msg *msg;
 	void *hdr;
-	int family, err = 0;
+	int err;
 
 	DBG("");
 
@@ -294,21 +285,21 @@ int __near_netlink_dep_link_up(uint32_t idx, uint32_t target_idx,
 	if (msg == NULL)
 		return -ENOMEM;
 
-	family = genl_family_get_id(nfc_state->nlnfc);
-
-	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, family, 0,
+	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, nfc_state->nfc_id, 0,
 			NLM_F_REQUEST, NFC_CMD_DEP_LINK_UP, NFC_GENL_VERSION);
 	if (hdr == NULL) {
 		err = -EINVAL;
 		goto nla_put_failure;
 	}
 
+	err = -EMSGSIZE;
+
 	NLA_PUT_U32(msg, NFC_ATTR_DEVICE_INDEX, idx);
 	NLA_PUT_U32(msg, NFC_ATTR_TARGET_INDEX, target_idx);
 	NLA_PUT_U8(msg, NFC_ATTR_COMM_MODE, comm_mode);
 	NLA_PUT_U8(msg, NFC_ATTR_RF_MODE, rf_mode);
 
-	err = nl_send_msg(nfc_state->nl_sock, msg, NULL, NULL);
+	err = nl_send_msg(nfc_state->cmd_sock, msg, NULL, NULL);
 
 nla_put_failure:
 	nlmsg_free(msg);
@@ -320,7 +311,7 @@ int __near_netlink_dep_link_down(uint32_t idx)
 {
 	struct nl_msg *msg;
 	void *hdr;
-	int family, err = 0;
+	int err;
 
 	DBG("");
 
@@ -328,18 +319,18 @@ int __near_netlink_dep_link_down(uint32_t idx)
 	if (msg == NULL)
 		return -ENOMEM;
 
-	family = genl_family_get_id(nfc_state->nlnfc);
-
-	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, family, 0,
+	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, nfc_state->nfc_id, 0,
 			NLM_F_REQUEST, NFC_CMD_DEP_LINK_DOWN, NFC_GENL_VERSION);
 	if (hdr == NULL) {
 		err = -EINVAL;
 		goto nla_put_failure;
 	}
 
+	err = -EMSGSIZE;
+
 	NLA_PUT_U32(msg, NFC_ATTR_DEVICE_INDEX, idx);
 
-	err = nl_send_msg(nfc_state->nl_sock, msg, NULL, NULL);
+	err = nl_send_msg(nfc_state->cmd_sock, msg, NULL, NULL);
 
 nla_put_failure:
 	nlmsg_free(msg);
@@ -351,7 +342,7 @@ int __near_netlink_adapter_enable(int idx, near_bool_t enable)
 {
 	struct nl_msg *msg;
 	void *hdr;
-	int family, err = 0;
+	int err;
 	uint8_t cmd;
 
 	DBG("");
@@ -360,23 +351,23 @@ int __near_netlink_adapter_enable(int idx, near_bool_t enable)
 	if (msg == NULL)
 		return -ENOMEM;
 
-	family = genl_family_get_id(nfc_state->nlnfc);
-
 	if (enable == TRUE)
 		cmd = NFC_CMD_DEV_UP;
 	else
 		cmd = NFC_CMD_DEV_DOWN;
 
-	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, family, 0,
+	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, nfc_state->nfc_id, 0,
 			NLM_F_REQUEST, cmd, NFC_GENL_VERSION);
 	if (hdr == NULL) {
 		err = -EINVAL;
 		goto nla_put_failure;
 	}
 
+	err = -EMSGSIZE;
+
 	NLA_PUT_U32(msg, NFC_ATTR_DEVICE_INDEX, idx);
 
-	err = nl_send_msg(nfc_state->nl_sock, msg, NULL, NULL);
+	err = nl_send_msg(nfc_state->cmd_sock, msg, NULL, NULL);
 
 nla_put_failure:
 	nlmsg_free(msg);
@@ -484,7 +475,7 @@ static int nfc_netlink_event_targets_found(struct genlmsghdr *gnlh)
 	struct nlattr *attr[NFC_ATTR_MAX + 1];
 	struct nl_msg *msg;
 	void *hdr;
-	int err, family;
+	int err;
 	uint32_t adapter_idx;
 
 	DBG("");
@@ -502,18 +493,18 @@ static int nfc_netlink_event_targets_found(struct genlmsghdr *gnlh)
 	if (msg == NULL)
 		return -ENOMEM;
 
-	family = genl_family_get_id(nfc_state->nlnfc);
-
-	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, family, 0,
+	hdr = genlmsg_put(msg, NL_AUTO_PID, NL_AUTO_SEQ, nfc_state->nfc_id, 0,
 			NLM_F_DUMP, NFC_CMD_GET_TARGET, NFC_GENL_VERSION);
 	if (hdr == NULL) {
 		err = -EINVAL;
 		goto nla_put_failure;
 	}
 
+	err = -EMSGSIZE;
+
 	NLA_PUT_U32(msg, NFC_ATTR_DEVICE_INDEX, adapter_idx);
 
-	err = nl_send_msg(nfc_state->nl_sock, msg,
+	err = nl_send_msg(nfc_state->cmd_sock, msg,
 				get_targets_handler, &adapter_idx);
 
 nla_put_failure:
@@ -607,6 +598,48 @@ static int nfc_netlink_event_dep_down(struct genlmsghdr *gnlh)
 	return 0;
 }
 
+static int nfc_netlink_event_tm_activated(struct genlmsghdr *gnlh)
+{
+	struct nlattr *attrs[NFC_ATTR_MAX + 1];
+	uint32_t idx;
+
+	DBG("");
+
+	nla_parse(attrs, NFC_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
+		  genlmsg_attrlen(gnlh, 0), NULL);
+	if (attrs[NFC_ATTR_DEVICE_INDEX] == NULL) {
+		near_error("Missing device index");
+		return -ENODEV;
+	}
+
+	idx = nla_get_u32(attrs[NFC_ATTR_DEVICE_INDEX]);
+
+	DBG("%d", idx);
+
+	return __near_adapter_add_device(idx, NULL, 0);
+}
+
+static int nfc_netlink_event_tm_deactivated(struct genlmsghdr *gnlh)
+{
+	struct nlattr *attrs[NFC_ATTR_MAX + 1];
+	uint32_t idx;
+
+	DBG("");
+
+	nla_parse(attrs, NFC_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
+		  genlmsg_attrlen(gnlh, 0), NULL);
+	if (attrs[NFC_ATTR_DEVICE_INDEX] == NULL) {
+		near_error("Missing device index");
+		return -ENODEV;
+	}
+
+	idx = nla_get_u32(attrs[NFC_ATTR_DEVICE_INDEX]);
+
+	DBG("%d", idx);
+
+	return __near_adapter_remove_device(idx);
+}
+
 static int nfc_netlink_event(struct nl_msg *n, void *arg)
 {
 	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(n));
@@ -641,6 +674,16 @@ static int nfc_netlink_event(struct nl_msg *n, void *arg)
 		DBG("DEP link is down");
 		nfc_netlink_event_dep_down(gnlh);
 
+		break;
+	case NFC_EVENT_TM_ACTIVATED:
+		DBG("Target mode activated");
+		nfc_netlink_event_tm_activated(gnlh);
+
+		break;
+	case NFC_EVENT_TM_DEACTIVATED:
+		DBG("Target mode deactivated");
+		nfc_netlink_event_tm_deactivated(gnlh);
+
 		break;
 	}
 
@@ -663,7 +706,9 @@ static gboolean __nfc_netlink_event(GIOChannel *channel,
 	nl_cb_set(cb, NL_CB_SEQ_CHECK, NL_CB_CUSTOM, no_seq_check, NULL);
 	nl_cb_set(cb, NL_CB_VALID, NL_CB_CUSTOM, nfc_netlink_event, data);
 
-	nl_recvmsgs(state->nl_sock, cb);
+	nl_recvmsgs(state->event_sock, cb);
+
+	nl_cb_put(cb);
 
 	return TRUE;
 }
@@ -672,7 +717,7 @@ static int nfc_event_listener(struct nlnfc_state *state)
 {
 	int sock;
 
-	sock = nl_socket_get_fd(state->nl_sock);
+	sock = nl_socket_get_fd(state->event_sock);
 	netlink_channel = g_io_channel_unix_new(sock);
 	g_io_channel_set_close_on_unref(netlink_channel, TRUE);
 
@@ -704,7 +749,7 @@ static int family_handler(struct nl_msg *msg, void *arg)
 	nla_parse(tb, CTRL_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
 		  genlmsg_attrlen(gnlh, 0), NULL);
 
-        if (!tb[CTRL_ATTR_MCAST_GROUPS])
+	if (!tb[CTRL_ATTR_MCAST_GROUPS])
 		return NL_SKIP;
 
 	nla_for_each_nested(mcgrp, tb[CTRL_ATTR_MCAST_GROUPS], rem_mcgrp) {
@@ -722,7 +767,7 @@ static int family_handler(struct nl_msg *msg, void *arg)
 		grp->id = nla_get_u32(tb_mcgrp[CTRL_ATTR_MCAST_GRP_ID]);
 		break;
 	}
-	
+
 	return NL_SKIP;
 }
 
@@ -730,7 +775,7 @@ static int nl_get_multicast_id(struct nl_sock *sock, const char *family,
 				const char *group)
 {
 	struct nl_msg *msg;
-	int err = 0, ctrlid;
+	int err, ctrlid;
 	struct handler_args grp = {
 		.group = group,
 		.id = -ENOENT,
@@ -744,8 +789,9 @@ static int nl_get_multicast_id(struct nl_sock *sock, const char *family,
 
 	ctrlid = genl_ctrl_resolve(sock, "nlctrl");
 
-        genlmsg_put(msg, 0, 0, ctrlid, 0,
-		    0, CTRL_CMD_GETFAMILY, 0);
+	genlmsg_put(msg, 0, 0, ctrlid, 0, 0, CTRL_CMD_GETFAMILY, 0);
+
+	err = -EMSGSIZE;
 
 	NLA_PUT_STRING(msg, CTRL_ATTR_FAMILY_NAME, family);
 
@@ -773,54 +819,61 @@ int __near_netlink_init(void)
 	if (nfc_state == NULL)
 		return -ENOMEM;
 
-	nfc_state->nl_sock = nl_socket_alloc();
-	if (nfc_state->nl_sock == NULL) {
-		near_error("Failed to allocate NFC netlink socket");
+	nfc_state->cmd_sock = nl_socket_alloc();
+	if (nfc_state->cmd_sock == NULL) {
+		near_error("Failed to allocate NFC command netlink socket");
 		err = -ENOMEM;
 		goto state_free;
 	}
 
-	if (genl_connect(nfc_state->nl_sock)) {
+	nfc_state->event_sock = nl_socket_alloc();
+	if (nfc_state->event_sock == NULL) {
+		near_error("Failed to allocate NFC event netlink socket");
+		err = -ENOMEM;
+		goto handle_cmd_destroy;
+	}
+
+	if (genl_connect(nfc_state->cmd_sock)) {
 		near_error("Failed to connect to generic netlink");
 		err = -ENOLINK;
-		goto handle_destroy;
+		goto handle_event_destroy;
 	}
 
-	if (genl_ctrl_alloc_cache(nfc_state->nl_sock, &nfc_state->nl_cache)) {
-		near_error("Failed to allocate generic netlink cache");
-		err = -ENOMEM;
-		goto handle_destroy;
+	if (genl_connect(nfc_state->event_sock)) {
+		near_error("Failed to connect to generic netlink");
+		err = -ENOLINK;
+		goto handle_event_destroy;
 	}
 
-	nfc_state->nlnfc = genl_ctrl_search_by_name(nfc_state->nl_cache, "nfc");
-	if (nfc_state->nlnfc == NULL) {
-		near_error("nfc not found");
+	nfc_state->nfc_id = genl_ctrl_resolve(nfc_state->cmd_sock, "nfc");
+	if (nfc_state->nfc_id < 0) {
+		near_error("Unable to find NFC netlink family");
 		err = -ENOENT;
-		goto cache_free;
+		goto handle_event_destroy;
 	}
 
-	nfc_state->mcid = nl_get_multicast_id(nfc_state->nl_sock, NFC_GENL_NAME,
+	nfc_state->mcid = nl_get_multicast_id(nfc_state->cmd_sock, NFC_GENL_NAME,
 						NFC_GENL_MCAST_EVENT_NAME);
 	if (nfc_state->mcid <= 0) {
 		near_error("Wrong mcast id %d", nfc_state->mcid);
 		err = nfc_state->mcid;
-		goto family_free;
+		goto handle_event_destroy;
 	}
 
-	err = nl_socket_add_membership(nfc_state->nl_sock, nfc_state->mcid);
+	err = nl_socket_add_membership(nfc_state->event_sock, nfc_state->mcid);
 	if (err) {
-		near_error("Error adding nl socket to membership");
-		goto family_free;
+		near_error("Error adding nl event socket to membership");
+		goto handle_event_destroy;
 	}
 
 	return nfc_event_listener(nfc_state);
 
-family_free:
-	genl_family_put(nfc_state->nlnfc);
-cache_free:
-	nl_cache_free(nfc_state->nl_cache);
-handle_destroy:
-	nl_socket_free(nfc_state->nl_sock);
+handle_event_destroy:
+	nl_socket_free(nfc_state->event_sock);
+
+handle_cmd_destroy:
+	nl_socket_free(nfc_state->cmd_sock);
+
 state_free:
 	g_free(nfc_state);
 
@@ -843,12 +896,10 @@ void __near_netlink_cleanup(void)
 	if (nfc_state == NULL)
 		return;
 
-	genl_family_put(nfc_state->nlnfc);
-	nl_cache_free(nfc_state->nl_cache);
-	nl_socket_free(nfc_state->nl_sock);
+	nl_socket_free(nfc_state->cmd_sock);
+	nl_socket_free(nfc_state->event_sock);
 
 	g_free(nfc_state);
 
 	DBG("");
 }
-
