@@ -63,6 +63,7 @@ struct near_tag {
 	struct {
 		uint8_t IDm[TYPE3_IDM_LEN];
 		uint8_t attr[TYPE3_ATTR_BLOCK_SIZE];
+		uint8_t ic_type;
 	} t3;
 
 	struct {
@@ -71,6 +72,7 @@ struct near_tag {
 	} t4;
 
 	DBusMessage *write_msg; /* Pending write message */
+	struct near_ndef_message *write_ndef;
 };
 
 static DBusConnection *connection = NULL;
@@ -226,6 +228,23 @@ static DBusMessage *set_property(DBusConnection *conn,
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
+static void tag_read_cb(uint32_t adapter_idx, uint32_t target_idx, int status)
+{
+	struct near_tag *tag;
+
+	tag = near_tag_get_tag(adapter_idx, target_idx);
+
+	if (tag == NULL)
+		return;
+
+	dbus_message_unref(tag->write_msg);
+	tag->write_msg = NULL;
+
+	__near_adapter_start_check_presence(adapter_idx, target_idx);
+
+	__near_adapter_tags_changed(adapter_idx);
+}
+
 static void write_cb(uint32_t adapter_idx, uint32_t target_idx, int status)
 {
 	struct near_tag *tag;
@@ -234,12 +253,14 @@ static void write_cb(uint32_t adapter_idx, uint32_t target_idx, int status)
 
 	DBG("Write status %d", status);
 
-	conn = near_dbus_get_connection();
 	tag = near_tag_get_tag(adapter_idx, target_idx);
-
-	if (conn == NULL || tag == NULL)
+	if (tag == NULL)
 		return;
 
+	conn = near_dbus_get_connection();
+	if (conn == NULL)
+		goto out;
+
 	if (status != 0) {
 		reply = __near_error_failed(tag->write_msg, EINVAL);
 		if (reply != NULL)
@@ -248,8 +269,56 @@ static void write_cb(uint32_t adapter_idx, uint32_t target_idx, int status)
 		g_dbus_send_reply(conn, tag->write_msg, DBUS_TYPE_INVALID);
 	}
 
+	near_ndef_records_free(tag->records);
+	tag->n_records = 0;
+	tag->records = NULL;
+	g_free(tag->data);
+	tag->data = NULL;
+
+	if (status == 0) {
+		/*
+		 * If writing succeeded,
+		 * check presence will be restored after reading
+		 */
+		__near_tag_read(tag, tag_read_cb);
+		return;
+	}
+
+out:
 	dbus_message_unref(tag->write_msg);
 	tag->write_msg = NULL;
+
+	__near_adapter_start_check_presence(tag->adapter_idx, tag->target_idx);
+}
+
+static void format_cb(uint32_t adapter_idx, uint32_t target_idx, int status)
+{
+	struct near_tag *tag;
+	int err;
+
+	DBG("format status %d", status);
+
+	tag = near_tag_get_tag(adapter_idx, target_idx);
+	if (tag == NULL)
+		return;
+
+	if (tag->write_msg == NULL)
+		return;
+
+	if (status == 0) {
+		err = __near_tag_write(tag, tag->write_ndef,
+						write_cb);
+		if (err < 0)
+			goto error;
+	} else {
+		err = status;
+		goto error;
+	}
+
+	return;
+
+error:
+	write_cb(tag->adapter_idx, tag->target_idx, err);
 }
 
 static DBusMessage *write_ndef(DBusConnection *conn,
@@ -257,7 +326,7 @@ static DBusMessage *write_ndef(DBusConnection *conn,
 {
 	struct near_tag *tag = data;
 	struct near_ndef_message *ndef, *ndef_with_header = NULL;
-	int err;
+	int tlv_len_size, err;
 
 	DBG("conn %p", conn);
 
@@ -279,22 +348,38 @@ static DBusMessage *write_ndef(DBusConnection *conn,
 	switch (tag->type) {
 	case NFC_PROTO_JEWEL:
 	case NFC_PROTO_MIFARE:
+		if (ndef->length < 0xff)
+			tlv_len_size = 3;
+		else
+			tlv_len_size = 5;
+
 		ndef_with_header = g_try_malloc0(sizeof(
 					struct near_ndef_message));
 		if (ndef_with_header == NULL)
 			goto fail;
 
 		ndef_with_header->offset = 0;
-		ndef_with_header->length = ndef->length + 3;
-		ndef_with_header->data = g_try_malloc0(ndef->length + 3);
+		ndef_with_header->length = ndef->length + tlv_len_size;
+		ndef_with_header->data =
+				g_try_malloc0(ndef->length + tlv_len_size);
 		if (ndef_with_header->data == NULL)
 			goto fail;
 
 		ndef_with_header->data[0] = TLV_NDEF;
-		ndef_with_header->data[1] = ndef->length;
-		memcpy(ndef_with_header->data + 2, ndef->data, ndef->length);
-		ndef_with_header->data[ndef->length + 2] = TLV_END;
 
+		if (ndef->length < 0xff) {
+			ndef_with_header->data[1] = ndef->length;
+		} else {
+			ndef_with_header->data[1] = 0xff;
+			ndef_with_header->data[2] =
+					(uint8_t)(ndef->length >> 8);
+			ndef_with_header->data[3] = (uint8_t)(ndef->length);
+		}
+
+		memcpy(ndef_with_header->data + tlv_len_size - 1, ndef->data,
+				ndef->length);
+		ndef_with_header->data[ndef->length + tlv_len_size - 1] =
+									TLV_END;
 		break;
 
 	case NFC_PROTO_FELICA:
@@ -342,6 +427,7 @@ static DBusMessage *write_ndef(DBusConnection *conn,
 	g_free(ndef->data);
 	g_free(ndef);
 
+	tag->write_ndef = ndef_with_header;
 	err = __near_tag_write(tag, ndef_with_header, write_cb);
 	if (err < 0) {
 		g_free(ndef_with_header->data);
@@ -412,8 +498,8 @@ void __near_tag_append_records(struct near_tag *tag, DBusMessageIter *iter)
 
 static enum near_tag_sub_type get_tag_type2_sub_type(uint8_t sel_res)
 {
-	switch(sel_res) {
-	case 0x00 :
+	switch (sel_res) {
+	case 0x00:
 		return NEAR_TAG_NFC_T2_MIFARE_ULTRALIGHT;
 	case 0x08:
 		return NEAR_TAG_NFC_T2_MIFARE_CLASSIC_1K;
@@ -423,7 +509,7 @@ static enum near_tag_sub_type get_tag_type2_sub_type(uint8_t sel_res)
 		return NEAR_TAG_NFC_T2_MIFARE_CLASSIC_4K;
 	case 0x20:
 		return NEAR_TAG_NFC_T2_MIFARE_DESFIRE;
-	case 0x28 :
+	case 0x28:
 		return NEAR_TAG_NFC_T2_JCOP30;
 	case 0x38:
 		return NEAR_TAG_NFC_T2_MIFARE_4K_EMUL;
@@ -462,7 +548,7 @@ static void set_tag_type(struct near_tag *tag,
 
 		DBG("proto 0x%x", proto);
 
-		switch(proto) {
+		switch (proto) {
 		case NFC_TAG_A_TYPE2:
 			tag->type = NFC_PROTO_MIFARE;
 			tag->sub_type = get_tag_type2_sub_type(sel_res);
@@ -578,12 +664,6 @@ const char *__near_tag_get_path(struct near_tag *tag)
 	return tag->path;
 }
 
-
-uint32_t __near_tag_get_idx(struct near_tag *tag)
-{
-	return tag->target_idx;
-}
-
 uint32_t __near_tag_get_type(struct near_tag *tag)
 {
 	return tag->type;
@@ -694,6 +774,8 @@ int near_tag_add_records(struct near_tag *tag, GList *records,
 		tag->records = g_list_append(tag->records, record);
 	}
 
+	__near_agent_ndef_parse_records(tag->records);
+
 	if (cb != NULL)
 		cb(tag->adapter_idx, tag->target_idx, status);
 
@@ -712,6 +794,11 @@ void near_tag_set_blank(struct near_tag *tag, near_bool_t blank)
 	tag->blank = blank;
 }
 
+near_bool_t near_tag_get_blank(struct near_tag *tag)
+{
+	return tag->blank;
+}
+
 uint8_t *near_tag_get_data(struct near_tag *tag, size_t *data_length)
 {
 	if (data_length == NULL)
@@ -817,6 +904,22 @@ uint8_t *near_tag_get_attr_block(struct near_tag *tag, uint8_t *len)
 	return tag->t3.attr;
 }
 
+void near_tag_set_ic_type(struct near_tag *tag, uint8_t ic_type)
+{
+	if (tag == NULL)
+		return;
+
+	tag->t3.ic_type = ic_type;
+}
+
+uint8_t near_tag_get_ic_type(struct near_tag *tag)
+{
+	if (tag == NULL)
+		return 0;
+
+	return tag->t3.ic_type;
+}
+
 static gint cmp_prio(gconstpointer a, gconstpointer b)
 {
 	const struct near_tag_driver *driver1 = a;
@@ -850,6 +953,9 @@ int __near_tag_read(struct near_tag *tag, near_tag_io_cb cb)
 
 	DBG("type 0x%x", tag->type);
 
+	/* Stop check presence while reading */
+	__near_adapter_stop_check_presence(tag->adapter_idx, tag->target_idx);
+
 	for (list = driver_list; list; list = list->next) {
 		struct near_tag_driver *driver = list->data;
 
@@ -878,22 +984,32 @@ int __near_tag_write(struct near_tag *tag,
 		DBG("driver type 0x%x", driver->type);
 
 		if (driver->type == tag->type) {
+			/* Stop check presence while writing */
+			__near_adapter_stop_check_presence(tag->adapter_idx,
+								tag->target_idx);
+
 			if (tag->blank == TRUE && driver->format != NULL) {
 				DBG("Blank tag detected, formatting");
 				err = driver->format(tag->adapter_idx,
-						tag->target_idx, NULL);
-
-				if (err < 0)
-					return err;
-
+						tag->target_idx, format_cb);
+			} else {
+				err = driver->write(tag->adapter_idx,
+							tag->target_idx, ndef,
+							cb);
 			}
 
-			return driver->write(tag->adapter_idx, tag->target_idx,
-								ndef, cb);
+			break;
 		}
 	}
 
-	return 0;
+	if (list == NULL)
+		err = -EOPNOTSUPP;
+
+	if (err < 0)
+		__near_adapter_start_check_presence(tag->adapter_idx,
+							tag->target_idx);
+
+	return err;
 }
 
 int __near_tag_check_presence(struct near_tag *tag, near_tag_io_cb cb)
